
//////////////////////////
#version 150
#pragma optimize(on)
#pragma debug(off)
#extension all : disable
//////////////////////////

// -- Constants --
const int MAX_NMB_LIGHTS = 8;
const int SHADING_FLAT = 0;
const int SHADING_PHONG = 1;
const vec2 SM_OFFSET = vec2(1.0 / 1024.0, 1.0 / 1024.0);

// -- Uniform --
uniform vec3 Camera_Position;

uniform int Shading_Type;
uniform int Shadow_Receiver;
uniform vec4 Color_Ambient;
uniform vec4 Color_Diffuse;
uniform vec4 Color_Specular;
uniform float Color_Specularity;

uniform sampler2D Tex_Ambient;
uniform sampler2D Tex_Diffuse;
uniform sampler2D Tex_Specular;
uniform sampler2D Tex_Specularity;
uniform sampler2D Tex_NormalMap;

uniform int Tex_Usage[5];

struct Light
{
	vec3 Position;
	vec3 Direction;
	vec4 ColorAmbient;
	vec4 ColorDiffuse;
	vec4 ColorSpecular;
	int IsShadowCaster;
	int Nmb_ShadowMaps;
	sampler2D Tex_ShadowMap1;
	sampler2D Tex_ShadowMap2;
	sampler2D Tex_ShadowMap3;
	mat4 Mat_LightProjView1;
	mat4 Mat_LightProjView2;
	mat4 Mat_LightProjView3;
	float ShadowFarDist[3];
};


uniform Light Lights[MAX_NMB_LIGHTS];

uniform int Nmb_LightIndices;
uniform int Light_Indices[MAX_NMB_LIGHTS];


// -- Input fragment --

in vec4 fPosWorld;
in vec4 fPosView;
in vec2 fTexCoord0;
in vec3 fNormalWorld;
in vec3 fTangentWorld;

// -- Fragment output --

out vec4 outColor;


vec3 GetNormal(sampler2D normalMap, vec3 normal, vec3 tangent, vec2 texcoord)
{
    vec3 normalTex = texture(normalMap, texcoord.xy).rgb;
    normalTex = 2.0f * normalTex - 1.0f;
	
	// Orthogonalize
	tangent =  normalize(tangent - dot(normal, tangent) * normal);
    vec3 bitangent = cross(normal, tangent);
	mat3 matTBN = mat3x3(tangent, bitangent, normal);
	return matTBN * normalTex;
}

int GetShadowMapIndex(int lightId) 
{
	int index = 0;
	Light light = Lights[lightId];
	for (int i = 0; i < light.Nmb_ShadowMaps; i++) {
		if (abs(fPosView.z) < light.ShadowFarDist[i]) {
			index = i;
			break;
		}
	}
	return index;
}

float ShadowLookupPCF(sampler2D shadowMap, vec3 smTexCoord, vec2 offset, float bias)
{
	if (texture(shadowMap, smTexCoord.xy + offset * SM_OFFSET).x + bias < smTexCoord.z) {
		return 0.0;
	}
	else {
		return 1.0;
	}
}

float CalculateShadowFactor(int lightId, vec3 normal)
{
	Light light = Lights[lightId];
	
	int smIndex = GetShadowMapIndex(lightId);
	float bias = 0.0;
	
	float shadowFactor = 1;
	if (light.IsShadowCaster == 1 && Shadow_Receiver == 1) {
	
		// Shadow map texture
		vec4 projCoordSM;
		if (smIndex == 0) {
			projCoordSM = light.Mat_LightProjView1 * fPosWorld;
		}
		else if (smIndex == 1) {
			projCoordSM = light.Mat_LightProjView2 * fPosWorld;
		}
		else if (smIndex == 2) {
			projCoordSM = light.Mat_LightProjView3 * fPosWorld;
		}
				
		vec3 smTexCoord = projCoordSM.xyz / projCoordSM.w;

		// PCF Filtering
		float sum = 0;
		float x, y;
		for (y = -1.5; y <= 1.5; y += 1.0) {
			for (x = -1.5; x <= 1.5; x += 1.0) {
				if (smIndex == 0) {
					bias = 0.001;
					sum += ShadowLookupPCF(light.Tex_ShadowMap1, smTexCoord, vec2(x, y), bias);
				}
				else if (smIndex == 1) {
					bias = 0.0015;
					sum += ShadowLookupPCF(light.Tex_ShadowMap2, smTexCoord, vec2(x, y), bias);
				}
				else {
					bias = 0.0015;
					sum += ShadowLookupPCF(light.Tex_ShadowMap3, smTexCoord, vec2(x, y), bias);
				}
			}
		}
		shadowFactor = sum / 16.0;
		shadowFactor = max(shadowFactor, 0.2);
		
	/***
		// Look up into shadow map texture
		if (smIndex == 0) {
			shadowFactor = texture(light.Tex_ShadowMap1, smTexCoord.xy).x;
		}
		else if (smIndex == 1) {
			shadowFactor = texture(light.Tex_ShadowMap2, smTexCoord.xy).x;
		}
		else {
			shadowFactor = texture(light.Tex_ShadowMap3, smTexCoord.xy).x;
		}
		
		if (shadowFactor + bias < smTexCoord.z) {
			shadowFactor = 0.2f;
		}
		else {
			shadowFactor = 1.0f;
		}
	***/
	}
	
	return shadowFactor;
}

vec4 FlatLight(int lightId, vec4 colorAmbient, vec4 colorDiffuse, vec3 normal)
{
	Light light = Lights[lightId];
	
	vec4 color = light.ColorAmbient * colorAmbient;
	color += light.ColorDiffuse * colorDiffuse;

    return color * CalculateShadowFactor(0, normal);
}

// Calculate color for a parallel light source
vec4 ParallelLight(int lightId, vec4 colorAmbient, vec4 colorDiffuse, vec4 colorSpecular, 
                   float specularity, vec3 normal)
{
	Light light = Lights[lightId];
	
	// Calculate shadow factor
	float shadowFactor = CalculateShadowFactor(0, normal);
	
	// Ambient term
	vec4 color = light.ColorAmbient * colorAmbient;
	
	// Diffuse and Specular term
	float kd = dot(-light.Direction, normal);
	if (kd > 0.0) {
		vec3 vecView = normalize(Camera_Position - fPosWorld.xyz);
		vec3 vecRefl = reflect(light.Direction, normal);
		float ks = pow(max(dot(vecRefl, vecView), 0.0), specularity);
		
		color += light.ColorDiffuse * colorDiffuse * kd;
		color += light.ColorSpecular * colorSpecular * ks;
	}
	
	return color * shadowFactor;
}



// -- Main --

void main()
{
	// Init color values	
	vec4 colorAmbient = Color_Ambient;
	vec4 colorDiffuse = Color_Diffuse;
	vec4 colorSpecular = Color_Specular;
	float colorSpecularity = Color_Specularity;
	vec3 normal = normalize(fNormalWorld);
	vec3 tangent = normalize(fTangentWorld);
	
	// Set color values from textures
	if( Tex_Usage[0] == 1) {
		colorAmbient = texture(Tex_Ambient, fTexCoord0.xy);
	}
	if( Tex_Usage[1] == 1) {
		colorDiffuse = texture(Tex_Diffuse, fTexCoord0.xy);
	}
	if( Tex_Usage[2] == 1) {
		colorSpecular = texture(Tex_Specular, fTexCoord0.xy);
	}
	if( Tex_Usage[3] == 1) {
		colorSpecularity = texture(Tex_Specularity, fTexCoord0.xy).r;
	}
	if( Tex_Usage[4] == 1) {
		normal = GetNormal(Tex_NormalMap, normal, tangent, fTexCoord0.xy);
	}

	if (Shading_Type == SHADING_FLAT) {
		// Flat Shading
		
		// For every light sum up the contribution
		for (int i = 0; i < Nmb_LightIndices; i++) {
			outColor = FlatLight(Light_Indices[i], colorAmbient, colorDiffuse, normal);
		}
	}
	else if (Shading_Type == SHADING_PHONG) {
		// Phong Shading
		
		// For every light sum up the contribution
		for (int i = 0; i < Nmb_LightIndices; i++) {
			outColor = ParallelLight(Light_Indices[i], colorAmbient, colorDiffuse, 
			                         colorSpecular, colorSpecularity, normal);
		}
	}
	else {
		Light light = Lights[0];
		//if (abs(fPosView.z) > light.ShadowFarDist[1]) {
		if(GetShadowMapIndex(0) == 0) {
			outColor = vec4(1, 0, 0, 1);
		}
		else if(GetShadowMapIndex(0) == 1) {
			outColor = vec4(0, 1, 0, 1);
		}
		else if(GetShadowMapIndex(0) == 2) {
			outColor = vec4(0, 0, 1, 1);
		}
		else {
			outColor = colorDiffuse;
		}
	}
}
